#include "game.h"
#include "level.h"
#include "frame.h"

//===== Player implementation =======//

num player::world_angle() const
{
    num arc = space_platform_arc();
    num angle = this->platform * arc;

    if(this->airborne()) {
        /* Time spent rotating; caps at the full rotation time */
        num t = num_clamp(this->jump_t, 0, TIME_ROTATION);
        angle += this->jump_dir * arc * t / TIME_ROTATION;
    }

    return angle;
}

vec2 player::world_angle_vector() const
{
    num angle = this->world_angle();
    return vec2(num_cos(angle), num_sin(angle));
}


vec3 player::pos() const
{
    /* Local position without accounting for current platform's angle */
    vec3 pos(0, -space_platform_distance() + this->height, this->z);

    return vec_rotate_around_z(pos, this->world_angle_vector());
}

vec3 player::forward() const
{
    return vec3(0, 0, 1);
}

vec3 player::backward() const
{
    return -this->forward();
}

vec3 player::up() const
{
    return vec_rotate_around_z(vec3(0, 1, 0), this->world_angle_vector());
}

vec3 player::down() const
{
    return -this->up();
}

void player::get_anim(struct anim **anim, int *frame)
{
    extern struct anim anim_erik_running;
    extern struct anim anim_erik_jumping;

    if(this->stance == Running) {
        this->frame = (this->frame + 1) % anim_erik_running.frame_count;
        *anim = &anim_erik_running;
        *frame = this->frame;
    }
    else {
        this->frame++;
        if(this->frame >= anim_erik_jumping.frame_count)
            this->frame = 12;
        *anim = &anim_erik_jumping;
        *frame = this->frame;
    }
}

//======= Functions on the game world space =======//

static vec2 platform_rotations[PLATFORM_COUNT];
static num platform_distance;

num space_platform_distance(void)
{
    return platform_distance;
}

/* We can't use a constructor function because g++ already generates a call to
   the default constructor, which would run after us and override the data. */
void space_init(void)
{
    num arc = space_platform_arc();

    for(int i = 0; i < PLATFORM_COUNT; i++) {
        num angle = arc * i - (arc / 2);
        platform_rotations[i] = vec2(num_cos(angle), num_sin(angle));
    }

    platform_distance = LEVEL_RADIUS * num_cos(arc / 2);
}

num space_platform_dist_from_center(int height_value)
{
    return LEVEL_RADIUS * (num(12 - height_value) / 12);
}

num space_platform_height(int height_value)
{
    return LEVEL_RADIUS - space_platform_dist_from_center(height_value);
}

struct prect space_platform_position(struct platform const *p)
{
    struct prect r;
    vec3 radius(0, -space_platform_dist_from_center(p->height), p->z);

    int angle_l = p->face;
    int angle_r = (p->face + 1) % PLATFORM_COUNT;

    r.nl = r.fl = vec_rotate_around_z(radius, platform_rotations[angle_l]);
    r.nr = r.fr = vec_rotate_around_z(radius, platform_rotations[angle_r]);

    r.fl.z += p->length;
    r.fr.z += p->length;
    return r;
}

//======= Game implementation =======//

void game_advance(struct game *game)
{
    struct level *level = &game->level;

    auto it = level->platform_buffer.begin();
    while(it != level->platform_buffer.end()) {
        /* If a platform is dead, remove it. We include the camera's distance
           in the calculation because we only remove platforms that have been
           passed *and* are now invisible. */
        if(game->player.z - RENDER_CAMERA_BACK_DISTANCE > it->z + it->length)
            it = level->platform_buffer.erase(it);
        else
            ++it;
    }
}

num game_height_at(struct game *game, num z, int face)
{
    int max_h = -1;

    for(auto const &p: game->level.platform_buffer) {
        if(p.face == face && z >= p.z && z <= p.z + p.length)
            max_h = (max_h >= p.height) ? max_h : p.height;
    }

    return max_h >= 0
        ? space_platform_height(max_h)
        : -KILL_PLANE_RADIUS;
}

const struct platform *game_platform_under_player(struct game const *game,
    struct player const *player)
{
    if(player->stance != player::Running)
        return NULL;

    for(auto const &p: game->level.platform_buffer) {
        if(player->platform == p.face
           && player->z >= p.z && player->z <= p.z + p.length) {
            return &p;
        }
    }

    return NULL;
}
